import os
import sys
from automated_ml_pack.modules.exception import CustomException
from automated_ml_pack.modules.logger import logging
import pandas as pd

from sklearn.model_selection import train_test_split
from dataclasses import dataclass

@dataclass
class DataIngestionConfig:
    basename_tag = "*"
    train_data_path: str=os.path.join('outputs',f"{basename_tag}train.csv")
    test_data_path: str=os.path.join('outputs',f"{basename_tag}test.csv")
    raw_data_path: str=os.path.join('outputs',f"{basename_tag}data.csv")

class DataIngestion:
    def __init__(self, 
                 input_data = None,
                 test_size:float = 0.2,
                 modeling_type:str = "reg",
                 target_column_name:str ="",
                 output_base:str ="",
                 output_dir:str = "outputs"
                 ):
        
        self.ingestion_config=DataIngestionConfig()
        basename_tag = self.ingestion_config.basename_tag
        self.ingestion_config.train_data_path = self.ingestion_config.train_data_path.replace(basename_tag,f"{modeling_type}_{output_base}_")
        self.ingestion_config.test_data_path = self.ingestion_config.test_data_path.replace(basename_tag,f"{modeling_type}_{output_base}_")
        self.ingestion_config.raw_data_path = self.ingestion_config.raw_data_path.replace(basename_tag,f"{modeling_type}_{output_base}_")

        self.ingestion_config.train_data_path = self.ingestion_config.train_data_path.replace("outputs",f"{output_dir}")
        self.ingestion_config.test_data_path = self.ingestion_config.test_data_path.replace("outputs",f"{output_dir}")
        self.ingestion_config.raw_data_path = self.ingestion_config.raw_data_path.replace("outputs",f"{output_dir}")

        self.input_data = input_data
        self.test_size = test_size
        self.target_column_name = target_column_name
        self.modeling_type = modeling_type
        self.output_base = output_base
        

    def initiate_data_ingestion(self):
        logging.info("Entered the data ingestion method or component")
        try:
            df=self.input_data
            logging.info('Successfully ingested dataset!')

            os.makedirs(os.path.dirname(self.ingestion_config.train_data_path),exist_ok=True)

            df.to_csv(self.ingestion_config.raw_data_path,index=False,header=True)
            df = df[df[self.target_column_name].notna()].reset_index(drop=True) # remove entries where the target value is missing
            
            logging.info("Train test split initiated")
            train_set,test_set=train_test_split(df,test_size=self.test_size,random_state=42)

            train_set.to_csv(self.ingestion_config.train_data_path,index=False,header=True)

            test_set.to_csv(self.ingestion_config.test_data_path,index=False,header=True)

            numerical_features = [feature for feature in df.columns if df[feature].dtype != 'O']
            categorical_features = [feature for feature in df.columns if df[feature].dtype == 'O']
            logging.info("Ingestion of the data is completed")

            return(
                self.ingestion_config.train_data_path,
                self.ingestion_config.test_data_path,
                numerical_features,
                categorical_features

            )
        except Exception as e:
            raise CustomException(e,sys)
        



import logging
import os
from datetime import datetime

LOG_FILE=f"{datetime.now().strftime('%m_%d_%Y_%H_%M_%S')}.log"
logs_path=os.path.join(os.getcwd(),"logs",LOG_FILE)
os.makedirs(logs_path,exist_ok=True)

LOG_FILE_PATH=os.path.join(logs_path,LOG_FILE)

logging.basicConfig(
    filename=LOG_FILE_PATH,
    format="[ %(asctime)s ] %(lineno)d %(name)s - %(levelname)s - %(message)s",
    level=logging.INFO,


)



import os
import sys
from dataclasses import dataclass

from catboost import CatBoostRegressor, CatBoostClassifier
from sklearn.ensemble import (
    AdaBoostRegressor, AdaBoostClassifier,
    GradientBoostingRegressor, GradientBoostingClassifier,
    RandomForestRegressor, RandomForestClassifier
)
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.metrics import r2_score, accuracy_score
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from xgboost import XGBRegressor, XGBClassifier 
from sklearn.svm import SVC
from sklearn.neural_network import MLPClassifier

from automated_ml_pack.modules.exception import CustomException
from automated_ml_pack.modules.logger import logging

from automated_ml_pack.modules.utils import save_object,getparams,evaluate_models, load_json, save_json

@dataclass
class ModelTrainerConfig:
    basename_tag = "*"
    trained_model_file_path=os.path.join("outputs",f"{basename_tag}model.pkl")
    model_report_path = os.path.join("outputs",f"{basename_tag}report.json")

class ModelTrainer:
    def __init__(self,
                 modeling_type = "reg",
                 output_base:str = "",
                 output_dir:str="outputs"
                 ):
        self.model_trainer_config=ModelTrainerConfig()
        self.basename_tag = self.model_trainer_config.basename_tag
        self.output_base = output_base
        self.model_trainer_config.model_report_path = self.model_trainer_config.model_report_path.replace(self.basename_tag,f"{modeling_type}_{output_base}_")
        self.model_trainer_config.model_report_path = self.model_trainer_config.model_report_path.replace("outputs",f"{output_dir}")
        self.modeling_type = modeling_type
        self.output_dir = output_dir


    def initiate_model_trainer(self,train_array,test_array):
        try:
            logging.info("Split training and test input data")
            X_train,y_train,X_test,y_test=(
                train_array[:,:-1],
                train_array[:,-1],
                test_array[:,:-1],
                test_array[:,-1]
            )

            if self.modeling_type == "reg":
                models = {
                    "Random Forest": RandomForestRegressor(random_state = 32),
                    "Decision Tree": DecisionTreeRegressor(random_state = 32),
                    "Gradient Boosting": GradientBoostingRegressor(random_state = 32),
                    "Linear Regression": LinearRegression(),
                    "XGBRegressor": XGBRegressor(random_state = 32),
                    "CatBoosting Regressor": CatBoostRegressor(random_state = 32,verbose=False),
                    "AdaBoost Regressor": AdaBoostRegressor(random_state = 32),
                }
            else:
                models = {
                    "Random Forest": RandomForestClassifier(random_state = 32),
                    "Decision Tree": DecisionTreeClassifier(random_state = 32),
                    "Gradient Boosting": GradientBoostingClassifier(random_state = 32),
                    "Logistic Regression": LogisticRegression(random_state = 32),
                    "XGBClassifier": XGBClassifier(random_state = 32),
                    "CatBoosting Classifier": CatBoostClassifier(random_state = 32,verbose=False),
                    "AdaBoost Classifier": AdaBoostClassifier(random_state = 32),
                    "MLPClassifier": MLPClassifier(random_state = 32, verbose=False),
                    "SVC": SVC(random_state = 32)
                }
            load_params = getparams()
            params = load_params["model_params"][self.modeling_type]

            logging.info("Training and Evaluating Models!")
            
            model_report:dict=evaluate_models(X_train=X_train,
                                              y_train=y_train,
                                              X_test=X_test,
                                              y_test=y_test,
                                              models=models,
                                              param=params,
                                              modeling_type = self.modeling_type
                                              )
            
            ## To get best model score from dict
            best_model_score = 0
            best_model_name = ""

            for model_name in model_report:
                if self.modeling_type == "reg":
                    score = model_report[model_name]["cross_val_report"]["r2"]["mean"]
                else:
                    score = model_report[model_name]["cross_val_report"]["accuracy"]["mean"]

                if score >= best_model_score:
                    best_model_score = score
                    best_model_name = model_name
            
            best_model = models[best_model_name]

            if best_model_score<0.5:
                raise "No best model found"
            logging.info(f"Best found model on both training and testing dataset")

            # save best model
            model_trainer_obj_file_path = self.model_trainer_config.trained_model_file_path.replace(self.basename_tag,f"{self.modeling_type}_{self.output_base}_{best_model_name}_")
            model_trainer_obj_file_path = model_trainer_obj_file_path.replace("outputs",f"{self.output_dir}")

            save_object(
                file_path=model_trainer_obj_file_path,
                obj=best_model
            )

            # save reports
            model_report_file_path = self.model_trainer_config.model_report_path
            save_json(model_report_file_path, model_report)

            return best_model_name, f"Best model: {best_model_name}\nBest model cross validation score: {best_model_score}"
            
            
        except Exception as e:
            raise CustomException(e,sys)



import argparse
import pandas as pd
import numpy as np
import os, sys
from dataclasses import dataclass
from sklearn.metrics import classification_report
import warnings
warnings.filterwarnings("ignore")

from automated_ml_pack.modules.logger import logging
from automated_ml_pack.modules.exception import CustomException

from automated_ml_pack.modules.components.data_transformation import DataTransformation
from automated_ml_pack.modules.components.data_ingestion import DataIngestion
from automated_ml_pack.modules.components.model_trainer import ModelTrainer
from automated_ml_pack.modules.utils import load_object, VisualizeResults

def main():
    # Parse command-line arguments
    parser = argparse.ArgumentParser(description='This tool facilitates the training of multiple machine learning models, optimizes the models, and saves the trained models. It also conducts model evaluation using diverse methods. Furthermore, the tool is capable of handling both regression and classification tasks. Additional options are described below.',
                                     usage='%(prog)s -[INPUT_FILE] [options]'
                                     )
    parser.add_argument('--input_file', required=True, type=str, help='Path to the input data in CSV/TSV format.')
    parser.add_argument('--input_type', type=str, choices=['csv', 'tsv'], default='csv', help='Type of input file format (csv or tsv)')
    parser.add_argument('--training_type', type=str, choices=['clf', 'reg'], default='clf', help='Type of training (e.g., "classification", "regression")')
    parser.add_argument('--target_column', type=str, required = True, help='Name of the target column in the input dataframe')
    parser.add_argument('--engineer_new_features', default=False, action="store_true",help='Flag to perform engineering of new features or not.')
    parser.add_argument('--output_base', type=str, default="", help='Base Name for most output files.')
    parser.add_argument('--test_size', type=float, default=0.2, help='What fraction of the dataset should be used for testing. Normally, cross validation is performed on the other percentage of the data to access the model\' generalization.')
    parser.add_argument('--no_standard_scaling', default = False, action="store_true", help='Whether or not to apply scikit-learn standard scaler on the data.')
    parser.add_argument('--feature_selection', default=False, action="store_true",help='Whether or not to perform feature selection on the dataset.')
    parser.add_argument('--feature_selection_method', type=str, default = "addition", choices = ["addition", "elimination"], help='Specify between recursive feature addition and recursive feature elimination algorithms for classification. By default, recursive feature addition is applied. For regression tasks, SelectKBest is used for feature selection.')
    parser.add_argument('--selectkbest_num_features', type=int, default = 32, help='Number of top features to select. For regression only.')
    parser.add_argument('--output_dir', type=str, default = "outputs", help='Custom Name of Output Folder.')
    parser.add_argument('--return_data', default = False, action="store_true", help='Select to include raw data, training data and test data in the output folders.')

    args = parser.parse_args()

    # Load dataframe from input file
    logging.info("Loading input data")
    try:
        if args.input_file:
            if args.input_type == 'csv':
                data = pd.read_csv(args.input_file)
            elif args.input_type == 'tsv':
                data = pd.read_csv(args.input_file, sep='\t')
            else:
                print("Invalid input type specified.")
                return
        else:
            raise ValueError("Please provide a valid input file path.")
            exit(1)
            
    except Exception as e:
        raise CustomException(e, sys)
    logging.info("Successfully Read the dataset as dataframe!")


    # Setting Input Configurations
    logging.info("Setting configurations based on provided parameters!")
    try:
        # Check if the target column exists in the dataframe
        if args.target_column not in data.columns:
            raise ValueError(f"Target column '{args.target_column}' not found in the input dataframe.")
            exit(1)

        # what kind of model to build
        if args.training_type =="clf":
            feature_selection_method = args.feature_selection_method
        elif args.training_type == "reg":
            feature_selection_method = "selectkbest"

        else:
            raise ValueError("--training_type argument must be either 'clf' or 'reg'!")
            exit(1)

        # create output folder
        os.makedirs(args.output_dir,exist_ok=True)

        engineer_new_features = args.engineer_new_features
        target_column = args.target_column
        training_type = args.training_type
        output_base = args.output_base
        test_size = args.test_size
        standard_scaling = not args.no_standard_scaling
        feature_selection = args.feature_selection
        selectkbest_num_features = args.selectkbest_num_features
        output_dir = args.output_dir
        return_data = args.return_data


    except Exception as e:
        raise CustomException(e, sys)
    logging.info("Completed setting configurations!")

    obj=DataIngestion(input_data = data, 
                      modeling_type = training_type,
                      test_size = test_size,
                      target_column_name = target_column,
                      output_base = output_base,
                      output_dir = output_dir
                      )
    train_data,test_data, numerical_columns, categorical_columns=obj.initiate_data_ingestion()
	
    logging.info("Transforming Data!")
    data_transformation=DataTransformation(numerical_columns=numerical_columns,
                                           categorical_columns = categorical_columns,
                                           target_column_name = target_column,
                                           modeling_type = training_type,
                                           engineer_new_features = engineer_new_features,
                                           standard_scaling = standard_scaling,
                                           output_base = output_base,
                                           feature_selection = feature_selection,
                                           feature_selection_method = feature_selection_method,
                                           selectkbest_num_features = selectkbest_num_features,
                                           output_dir = output_dir
                                           )
    train_arr,test_arr,_,selected_columns, reverse_encode=data_transformation.initiate_data_transformation(train_data,test_data)

    modeltrainer=ModelTrainer(modeling_type = training_type,
                              output_base = output_base,
                              output_dir = output_dir
                              )

    best_model_name, __ = modeltrainer.initiate_model_trainer(train_arr,test_arr)
    print(__)

    if training_type == "clf":
        # plot results
        logging.info("Plotting Results")
        results_plotter = VisualizeResults(os.path.join(f"{output_dir}",f"{training_type}_{output_base}_report.json"), output_dir = output_dir)
        results_plotter.run_make_plots()

        # get classification report for best model.
        test_data_path = os.path.join(f"{output_dir}",f"{training_type}_{output_base}_test.csv")
        test_df=pd.read_csv(test_data_path)
        preprocessing_obj=load_object(os.path.join(f"{output_dir}",f"{training_type}_{output_base}_preprocessor.pkl"))

        input_feature_test_df=test_df.drop(columns=[target_column],axis=1)
        target_feature_test_df=test_df[target_column].values

        input_feature_test_arr=preprocessing_obj.transform(input_feature_test_df)

        # load best model
        best_model_path = os.path.join(f"{output_dir}",f"{training_type}_{output_base}_{best_model_name}_model.pkl")
        best_model = load_object(best_model_path)
        y_pred = best_model.predict(input_feature_test_arr)
        y_pred = [reverse_encode[label] for label in y_pred]
        y_true = list(target_feature_test_df)
        target_names = [item for item in np.unique(y_pred)]

        clf_report = classification_report(y_true, y_pred)

        f = open(os.path.join(f"{output_dir}",f"{output_base}_classification_report.txt"), 'w')
        f.write(clf_report)
        f.close()

    # delete data from output folder
    if not return_data:
        os.remove(os.path.join(f"{output_dir}",f"{training_type}_{output_base}_data.csv"))
        os.remove(os.path.join(f"{output_dir}",f"{training_type}_{output_base}_train.csv"))
        os.remove(os.path.join(f"{output_dir}",f"{training_type}_{output_base}_test.csv"))

if __name__ == "__main__":
    main()



import sys
from dataclasses import dataclass

import numpy as np 
import pandas as pd
import os
from sklearn import set_config
set_config(transform_output = "pandas")

from feature_engine.creation import RelativeFeatures
from feature_engine.creation import MathFeatures
from feature_engine.selection import RecursiveFeatureElimination
from feature_engine.selection import RecursiveFeatureAddition
from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import f_regression
from sklearn.ensemble import RandomForestClassifier
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder,StandardScaler, LabelEncoder

from automated_ml_pack.modules.exception import CustomException
from automated_ml_pack.modules.logger import logging

from automated_ml_pack.modules.utils import save_object, save_json

@dataclass
class DataTransformationConfig:
    basename_tag = "*"
    preprocessor_obj_file_path=os.path.join('outputs',f"{basename_tag}preprocessor.pkl")
    selected_columns_path = os.path.join('outputs',f"{basename_tag}selected_columns.json")

class DataTransformation:
    def __init__(self, 
                 modeling_type:str = "reg",
                 categorical_columns:list = [],
                 numerical_columns:list = [],
                 target_column_name:str = "",
                 output_base:str = "",
                 engineer_new_features:bool = False,
                 standard_scaling:bool = True,
                 feature_selection:bool = False,
                 feature_selection_method:str = "addition",
                 selectkbest_num_features:int = 32,
                 output_dir:str="outputs"
                 ):
        """
        modeling_type can be either "clf" or "reg" for classification and regression repsectively.
        """
        self.data_transformation_config=DataTransformationConfig()
        basename_tag = self.data_transformation_config.basename_tag
        self.modeling_type = modeling_type
        self.output_base = output_base
        self.data_transformation_config.preprocessor_obj_file_path = self.data_transformation_config.preprocessor_obj_file_path.replace(basename_tag,f"{modeling_type}_{output_base}_")
        self.data_transformation_config.selected_columns_path = self.data_transformation_config.selected_columns_path.replace(basename_tag,f"{modeling_type}_{output_base}_")
        self.data_transformation_config.preprocessor_obj_file_path = self.data_transformation_config.preprocessor_obj_file_path.replace("outputs",f"{output_dir}")
        self.data_transformation_config.selected_columns_path = self.data_transformation_config.selected_columns_path.replace("outputs",f"{output_dir}")
        self.engineer_new_features = engineer_new_features
        self.standard_scaling = standard_scaling
        self.feature_selection = feature_selection
        self.feature_selection_method = feature_selection_method
        self.selectkbest_num_features = selectkbest_num_features
        self.target_column_name = target_column_name
        self.numerical_columns = numerical_columns
        self.categorical_columns = categorical_columns

        if target_column_name in self.categorical_columns:
            self.categorical_columns.remove(target_column_name)
        else:
            self.numerical_columns.remove(target_column_name)
            
            
    def get_data_transformer_object(self):
        '''
        This function si responsible for data trnasformation
        
        '''
        try:

            # set feature creation configurations
            if self.engineer_new_features:
                logging.info("Performing Feature Engineering!")
                mf_transformer = MathFeatures(variables = self.numerical_columns, func = ["sum", "mean", "median", "max", "min","std"])
                rf_num_transformer = RelativeFeatures(variables=self.numerical_columns,
                                                    reference=self.numerical_columns,
                                                    func = ["mul"])
            else:
                mf_transformer = None
                rf_num_transformer = None

            # set standard scaling configs
            if self.standard_scaling:
                logging.info("Performing Standard Scaling!")
                num_standard_scaler = StandardScaler()
                cat_standard_scaler = StandardScaler(with_mean=False)
            else:
                num_standard_scaler = None
                cat_standard_scaler = None

            # set configurations for feature selection and model training type
            feature_selection_methods = {
                "addition":RecursiveFeatureAddition(RandomForestClassifier(random_state=32), cv=3),
                "elimination": RecursiveFeatureElimination(RandomForestClassifier(random_state=32), cv=3),
                "selectkbest": SelectKBest(score_func=f_regression, k=self.selectkbest_num_features)
            }

            if self.feature_selection:
                logging.info("Performing Feature Selection!")
                feature_selector = feature_selection_methods[self.feature_selection_method]
            else:
                feature_selector = None

            

            num_pipeline= Pipeline(
                steps=[
                ("imputer",SimpleImputer(strategy="median")),
                ("math_features", mf_transformer),
                ("relative_features", rf_num_transformer),
                ("scaler",num_standard_scaler),
                ]
            )

            cat_pipeline=Pipeline(

                steps=[
                ("imputer",SimpleImputer(strategy="most_frequent")),
                ("one_hot_encoder",OneHotEncoder(sparse_output=False)),
                ("scaler",cat_standard_scaler)
                ]

            )

            preprocessor=ColumnTransformer(
                [
                ("num_pipeline",num_pipeline,self.numerical_columns),
                ("cat_pipelines",cat_pipeline,self.categorical_columns)

                ]
            )

            preprocessor_pipeline = Pipeline(
                steps = [
                    ("preprocessor", preprocessor),
                    ("feature_selection", feature_selector)
                ]
            )

            return preprocessor_pipeline
        
        except Exception as e:
            raise CustomException(e,sys)
        
    def initiate_data_transformation(self,train_path,test_path):

        try:
            train_df=pd.read_csv(train_path)
            test_df=pd.read_csv(test_path)

            logging.info("Read train and test data completed")

            logging.info("Obtaining preprocessing object")

            preprocessing_obj=self.get_data_transformer_object()

            input_feature_train_df=train_df.drop(columns=[self.target_column_name],axis=1)
            target_feature_train_df=train_df[self.target_column_name]

            input_feature_test_df=test_df.drop(columns=[self.target_column_name],axis=1)
            target_feature_test_df=test_df[self.target_column_name]

            logging.info(
                f"Applying preprocessing object on training dataframe and testing dataframe."
            )

            input_feature_train_arr=preprocessing_obj.fit_transform(input_feature_train_df, target_feature_train_df)
            selected_columns = list(input_feature_train_arr.columns)
            input_feature_test_arr=preprocessing_obj.transform(input_feature_test_df)
            
            if self.modeling_type == "clf":
                le = LabelEncoder()
                target_feature_train = le.fit_transform(target_feature_train_df.values)
                target_feature_test = le.transform(target_feature_test_df.values)
                reverse_encode = {le.transform([value])[0]:value for value in target_feature_test_df.unique()}
            else:
                target_feature_train = target_feature_train_df.values
                target_feature_test = target_feature_test_df.values
                reverse_encode = None

            train_arr = np.c_[
                input_feature_train_arr, target_feature_train
            ]
            test_arr = np.c_[input_feature_test_arr, target_feature_test]

            logging.info(f"Saved preprocessing object.")

            preprocessing_obj_file_path = self.data_transformation_config.preprocessor_obj_file_path
            save_object(

                file_path=preprocessing_obj_file_path,
                obj=preprocessing_obj

            )
	    
    	    # save selected column
            save_json(
                self.data_transformation_config.selected_columns_path,
                selected_columns
                )
	    
            return (
                train_arr,
                test_arr,
                self.data_transformation_config.preprocessor_obj_file_path,
                selected_columns,
                reverse_encode
            )
        except Exception as e:
            raise CustomException(e,sys)



import sys
from automated_ml_pack.modules.logger import logging

def error_message_detail(error,error_detail:sys):
    _,_,exc_tb=error_detail.exc_info()
    file_name=exc_tb.tb_frame.f_code.co_filename
    error_message="Error occured in python script name [{0}] line number [{1}] error message[{2}]".format(
     file_name,exc_tb.tb_lineno,str(error))

    return error_message

    

class CustomException(Exception):
    def __init__(self,error_message,error_detail:sys):
        super().__init__(error_message)
        self.error_message=error_message_detail(error_message,error_detail=error_detail)
    
    def __str__(self):
        return self.error_message



from setuptools import find_packages, setup
from typing import List

HYPHEN_E_DOT = "-e ."
def get_requirements(file_path) -> List[str]:
    """
    this function will return the list of requirements
    """

    requirements = []
    with open(file_path) as file_obj:
        requirements = file_obj.readlines()
        requirements = [package.replace("\n", "") for package in requirements]

        if HYPHEN_E_DOT in requirements:
            requirements.remove(HYPHEN_E_DOT)

    return requirements


with open("README.md", "r") as f:
    long_description = f.read()


setup(
name = "automated_ml_pack",
version= "1.0.3",
author="Cyrille",
author_email="cyrillemesue@gmail.com",
description='This package is designed for swift and automated machine learning practice, catering to both classification and regression tasks. It facilitates model training, grid search application, and the preservation of the best model. Furthermore, it stores and visualizes the best scores attained by other models using commonly employed evaluation metrics.',
package_dir={"": "src"},
packages=find_packages('src'),
long_description=long_description,
long_description_content_type="text/markdown",
url="https://github.com/CyrilleMesue/automated_ml_pack",
install_requires=get_requirements("requirements.txt"),
python_requires='>=3.11',
license="MIT",
classifiers=[
"License :: OSI Approved :: MIT License",
"Programming Language :: Python :: 3.11",
"Operating System :: OS Independent",
],
entry_points={
        'console_scripts': [
            'run_train_pipeline = automated_ml_pack.train_pipeline:main',
        ],
    },
)


import os
import sys

import numpy as np 
import pandas as pd
import seaborn as sns 
import matplotlib.pyplot as plt
from tqdm import tqdm
import json
import dill
import pickle
from sklearn.metrics import r2_score, mean_squared_error, mean_absolute_error, root_mean_squared_error
from sklearn.metrics import f1_score, accuracy_score, roc_auc_score, precision_score, recall_score
from sklearn.metrics import make_scorer
from sklearn.model_selection import cross_validate
from sklearn.model_selection import GridSearchCV, KFold, StratifiedKFold
from sklearn.metrics import classification_report

from automated_ml_pack.modules.exception import CustomException

def save_object(file_path, obj):
    try:
        dir_path = os.path.dirname(file_path)

        os.makedirs(dir_path, exist_ok=True)

        with open(file_path, "wb") as file_obj:
            pickle.dump(obj, file_obj)

    except Exception as e:
        raise CustomException(e, sys)

def evaluate_models(X_train, y_train,X_test,y_test,models,param, modeling_type = "reg"):
    try:
        report = {}

        if modeling_type =="reg":
            kfold = KFold(n_splits=5, random_state=32, shuffle = True)
            scoring = "r2"
        else:
            kfold = StratifiedKFold(n_splits=5, random_state=32, shuffle = True)
            scoring = "accuracy"
	
        for i in tqdm(range(len(list(models)))):
            model = list(models.values())[i]
            
            para=param[list(models.keys())[i]]

            gs = GridSearchCV(model,para,cv=kfold, scoring = scoring)
            gs.fit(X_train,y_train)

            model_best_params = gs.best_params_
            model.set_params(**model_best_params)
            model.fit(X_train,y_train)

            #get cross validation and test report
            cross_val_report = get_cross_validation_scores(model, X_train,y_train, cv= kfold, training_type = modeling_type)
            
            # make predictions
            y_train_pred = model.predict(X_train)
            y_test_pred = model.predict(X_test)

            train_report = get_test_report(y_train, y_train_pred, training_type=modeling_type)

            test_report = get_test_report(y_test, y_test_pred, training_type=modeling_type)

            report[list(models.keys())[i]] = {"test_report":test_report, 
                                              "train_report":train_report, 
                                              "cross_val_report":cross_val_report}

        return report

    except Exception as e:
        raise CustomException(e, sys)
    
def get_cross_validation_scores(model, X, y, cv, training_type = "clf"):
    """
    Get cross validation scores:
        ('f1', 'precision', 'recall', 'roc_auc', "accuracy") for classification
        ('r2', 'neg_mean_squared_error', 'neg_mean_absolute_error', 'neg_root_mean_squared_error') for regression
    """

    try:
        if training_type == "clf":
            scoring = ('f1', 'precision', 'recall', 'roc_auc', "accuracy")
        elif training_type == "reg":
            scoring = ('r2', 'neg_mean_squared_error', 'neg_mean_absolute_error', 'neg_root_mean_squared_error')
        else:
            print("Provide correct value for train_type: ['clf','reg']")

        scores = cross_validate(model, X, y, cv=cv,scoring=scoring,return_train_score=False)
        score_report = {"_".join(score_name.split("_")[1:]):{"mean":scores[score_name].mean(), 
                                                         "std":scores[score_name].std(),
                                                         "all":list(scores[score_name])} for score_name in scores}
    except Exception as e:
        raise CustomException(e, sys)
    
    return score_report

def get_test_report(true, predicted, training_type='clf'):

    """
    Run Various Evaluation Metrics on data
    """

    try:
        if training_type == 'clf':
            score_report = {"f1":f1_score(true, predicted),
                            "accuracy": accuracy_score(true, predicted),
                            "roc_auc": roc_auc_score(true, predicted),
                            "precision":precision_score(true, predicted),
                            "recall":recall_score(true, predicted)
                        }
            return score_report

        elif training_type == 'reg':
            score_report = {
                            "r2": r2_score(true, predicted),
                            "neg_mean_squared_error":-mean_squared_error(true, predicted),
                            "neg_mean_absolute_error":-mean_absolute_error(true, predicted),
                            "neg_root_mean_squared_error": -mean_squared_error(true, predicted, squared=False)
            }
            return score_report
        else:
            raise ValueError("Invalid training type. Choose 'clf' or 'reg'.")
    except Exception as e:
        raise CustomException(e, sys)
    
def load_object(file_path):
    try:
        with open(file_path, "rb") as file_obj:
            return pickle.load(file_obj)

    except Exception as e:
        raise CustomException(e, sys)
    

def load_json(json_path):
    """
    Load json data as a dictionary
    """

    try:
        with open(json_path) as f:
            file = f.read()
        json_data = json.loads(file)

    except Exception as e:
        raise CustomException(e, sys)

    return json_data


def save_json(file_path, obj):
    try:
        dir_path = os.path.dirname(file_path)

        os.makedirs(dir_path, exist_ok=True)

        json_object = json.dumps(obj, indent=2)

        with open(file_path, "w") as file_obj:
            file_obj.write(json_object)

    except Exception as e:
        raise CustomException(e, sys)



class VisualizeResults:
    """
    This class is used to visualized the results obtained from the machine learning pack. 
    """

    def __init__(self, results_path:str = None,
                 output_dir:str = "outputs"
                ):
        self.output_dir = output_dir
        # load results data as a dictionary
        self.results_data = load_json(results_path)
    
    def prepare_error_bar_data(self, report_type:str = "cross_val_report"):
        """
        convert results to dataframe format for errorbar plotting
        """

        try:
            reshaped_results_data = {"score":[], "model":[], "metric":[]}
            for model_name in self.results_data:
                for metric in self.results_data[model_name][report_type]:
                    if report_type == "cross_val_report":
                        score = self.results_data[model_name][report_type][metric]["mean"]
                    elif report_type == "train_report":
                        score = self.results_data[model_name][report_type][metric]
                    elif report_type == "test_report":
                        score = self.results_data[model_name][report_type][metric]
                    else:
                        raise ValueError("Please enter correct value for report_type as one of ['cross_val_report', 'train_report', 'test_report']")
                    reshaped_results_data["score"].append(score)
                    reshaped_results_data["model"].append(model_name)
                    reshaped_results_data["metric"].append(metric)
            
            reshaped_results_data = pd.DataFrame(reshaped_results_data)
            reshaped_results_data = reshaped_results_data[reshaped_results_data.metric !="time"]
            return reshaped_results_data
        except Exception as e:
            raise CustomException(e, sys)

    
    def prepare_barplot_data(self, report_type:str = "cross_val_report"):
        """
        convert results to dataframe format for barplot plotting
        """
        try:
            reshaped_results_data = {name:[] for name in self.results_data}
            reshaped_results_data["metric"] = []
            report_type = report_type
            metrics = self.results_data["Random Forest"][report_type].keys()
            
            for metric in metrics:
                reshaped_results_data["metric"].append(metric)
                for model_name in self.results_data:
                    if report_type == "cross_val_report":
                        score = self.results_data[model_name][report_type][metric]["mean"]
                    elif report_type == "train_report":
                        score = self.results_data[model_name][report_type][metric]
                    elif report_type == "test_report":
                        score = self.results_data[model_name][report_type][metric]
                    else:
                        raise ValueError("Please enter correct value for report_type as one of ['cross_val_report', 'train_report', 'test_report']")

                    reshaped_results_data[model_name].append(score)
                    
            reshaped_results_data = pd.DataFrame(reshaped_results_data)
            reshaped_results_data = reshaped_results_data[reshaped_results_data.metric != "time"]
            reshaped_results_data.rename(columns={'Random Forest': 'RandomForest',
                               'Decision Tree': 'Tree',
                               'Gradient Boosting': 'GradientBoost',
                               'Logistic Regression': 'Logistic Regression',
                               'XGBClassifier': 'XGBoost',
                               'CatBoosting Classifier': 'CatBoost',
                               'AdaBoost Classifier': 'AdaBoost',
                               'MLPClassifier': 'NN',
                               "Linear Regression": "Linear Regression",
                               "XGBRegressor": 'XGBoost',
                               "CatBoosting Regressor": 'CatBoost',
                               "AdaBoost Regressor": 'AdaBoost',
                               }, inplace=True)

        except Exception as e:
            raise CustomException(e, sys)
        return reshaped_results_data

    
    def plot_errorbars(self,
                       report_type:str = "cross_val_report",
                       score_column:str= "score", 
                       hue_colum:str = "model",
                       xlabel_column:str = "metric", 
                       xlabel:str = "Evaluation Metric", 
                       ylabel:str="Score"):
        """
        receive score data and plot error bar
        """

        try:
            data = self.prepare_error_bar_data(report_type = report_type)
            # set colors
            colors = ['green', 'lime', 'magenta','orange', 'blue', 'cyan', 'red', 'black' , 'gray','brown', 'purple', 'yellow','salmon','pimk']
            
            nrows = 1
            ncols = data[xlabel_column].nunique()
            figure, axes  = plt.subplots(nrows=1 , ncols = ncols , figsize = (20, 10)) 
        
            labels = list(data[xlabel_column].unique())
            hue = list(data[hue_colum].unique())
            n_hue = data[hue_colum].nunique()
        
            ## set the labels of the X-Axis
            for i , label in enumerate(labels):
                axes[i,].set(xlabel=label)
        
            for i in range(1, ncols):
              #set horizontal axis
              #axes[i,].set_ylim((0.5, 1.00))
              axes[i,].yaxis.grid(True)
              axes[i].set_yticklabels([])    
        
            #set horizontal axis of first column plot in the figure
            axes[0,].yaxis.grid()
            
            #set the y-axis label for the first plot
            axes[0,].set_ylabel(f'{ylabel}', fontsize='large', fontweight='bold')
        
            ##set the font-size of the x-axis and y-axis labels
            for i in range(ncols):
              axes[i,].xaxis.label.set_size(24)
              if i == 0:
                axes[0,].yaxis.label.set_size(30)
                  
            x = range(1, n_hue+1)
        
            scores = {label:{"scores":[], "error_rate":[]} for label in labels}
            # create score list
            for label in labels:
                for type in hue:
                    score = data[data[xlabel_column] == label][data[hue_colum]==type].values[0,0]
                    error = 1-score
                    scores[label]["scores"].append(score)
                    scores[label]["error_rate"].append(error)
        
            # plot graphs
            for i, label in enumerate(labels):
                axes[i,].scatter(x,scores[label]["scores"] , s= 100, color = colors[:n_hue])
                #plot the error bars
                for pos, y, err, color in zip(x, scores[label]["scores"],scores[label]["error_rate"] , colors[:n_hue]):
                    axes[i,].errorbar(pos, y, err, lw = 4,capsize = 4, capthick = 2 , color = color)  
                    
            i = 0
            for pos, y, err,color in zip(x, scores[label]["scores"],scores[label]["error_rate"],colors):
                 axes[ncols-1,].errorbar(pos, y, err, lw = 4, label=hue[i], capsize = 4, capthick = 2, color = color)
                 i = i + 1
            
            ##add common x-axis label to the Figure
            figure.text(0.5, 0.04, f'{xlabel}', ha='center', fontsize = 30, weight = 'bold')
        
            # change the fontsize of x and y axis ticks
            axes[0,].tick_params(axis='y', labelsize=24)
            
            for i in range(len(labels)):
              axes[i,].tick_params(axis = 'x', labelsize = 24)
              #remove x axis ticks and labels
              axes[i, ].tick_params(right = False, labelbottom = False, bottom = False)
            plt.legend(ncol=ncols ,loc="lower center", bbox_to_anchor=(-1.9, 1.02), fontsize = 17 , handlelength=0.5, handleheight=0.5) 
            plt.savefig(f"{self.output_dir}/{report_type}_error.png", bbox_inches='tight')
            plt.savefig(f"{self.output_dir}/{report_type}_error.pdf", bbox_inches='tight', format="pdf")

        except Exception as e:
            raise CustomException(e, sys)


    def barplots(self, 
                 report_type:str = "cross_val_report", 
                 x_column:str = "metric",
                 xlabel:str = "Evaluation Metric", 
                 ylabel:str="Score"
                ):
        """
        plot bar plots from score data.
        """
        try:
            result_data = self.prepare_barplot_data(report_type = "cross_val_report")
            colors = ['green', 'lime', 'magenta','orange', 'blue', 'cyan', 'red', 'black' , 'gray','brown', 'purple', 'yellow','salmon','pimk']
        
            n_cols = len(result_data.columns)-1
            # plot grouped bar chart
            result_data.plot(x=x_column,
                             kind='bar',
                             width=0.7,
                             stacked=False,
                             color = colors[:n_cols],
                             figsize=(20, 10))
            
            plt.xlabel(f'{xlabel}', fontsize = 30 , weight = 'bold');
            plt.ylabel(f'{ylabel}', fontsize=30, weight = 'bold');
            #plt.ylim((0.50, 1.00))
            plt.xticks(fontsize= 24 , rotation = 360)
            plt.yticks(fontsize= 24)
            axes = plt.gca()
            axes.yaxis.grid()
            plt.legend(ncol=n_cols, loc="lower center", bbox_to_anchor=(0.5, 1.02), fontsize = 17,handlelength=0.6, handleheight=0.6)
            plt.savefig(f"{self.output_dir}/{report_type}_bar.png", bbox_inches='tight')
            plt.savefig(f"{self.output_dir}/{report_type}_barr.pdf", bbox_inches='tight', format="pdf")
        except Exception as e:
            raise CustomException(e, sys)

    
    def run_make_plots(self):
        """
        make visualizations of ml results and save as pdf and png.
        """
    
        options = {"cross_val_report":"Cross Validation Score", "train_report":"Training Score", "test_report":"Test Score"}
        
        try:
            for option in options:
                    self.plot_errorbars(report_type = option, ylabel = options[option])
                    self.barplots(report_type = option, ylabel = options[option])
        except Exception as e:
            raise CustomException(e, sys)


def getparams():
    params = {"model_params" : 
    {"reg" : {"Decision Tree": 
            {"criterion": ["squared_error", "friedman_mse", "absolute_error", "poisson"]},
            "Random Forest": 
            {"n_estimators": [8, 16, 32, 64, 128, 256, 512]},
            "Gradient Boosting": 
            {"n_estimators": [8, 16, 32, 64, 128, 256, 512]},
            "Linear Regression": {},
            "XGBRegressor": {
                "learning_rate": [0.1, 0.01, 0.05, 0.001],
                "n_estimators": [8, 16, 32, 64, 128, 256, 512]
                },
            "CatBoosting Regressor": 
                {
                    "depth": [6, 8, 10],
                    "learning_rate": [0.01, 0.05, 0.1],
                    "iterations": [30, 50, 100]
                },
            "AdaBoost Regressor": 
                {
                    "learning_rate": [0.1, 0.01, 0.5, 0.001],
                    "n_estimators": [8, 16, 32, 64, 128, 256, 512]
                }
            },
        "clf":
            {"Decision Tree": 
                {
                    "criterion": ["gini", "entropy"]
                },
            "Random Forest": 
                {
                    "n_estimators": [8, 16, 32, 64, 128, 256, 512]
                },
            "Gradient Boosting": 
                {
                    "n_estimators": [8, 16, 32, 64, 128, 256, 512]
                },
            "Logistic Regression": {},
            "XGBClassifier": 
                {
                    "learning_rate": [0.1, 0.01, 0.05, 0.001],
                    "n_estimators": [8, 16, 32, 64, 128, 256, 512],
                    "max_depth": [3,5,8]
                },
            "CatBoosting Classifier": 
                {
                    "depth": [6, 8, 10],
                    "learning_rate": [0.01, 0.05, 0.1],
                    "iterations": [30, 50, 100]
                },
            "AdaBoost Classifier": 
                {
                    "learning_rate": [0.1, 0.01, 0.5, 0.001],
                    "n_estimators": [8, 16, 32, 64, 128, 256, 512]
                },
            "MLPClassifier":{},
            "SVC":
                {
                    "C": [0.1,1,10,100]
                }
            }

        }
    }
    return params